/**
 * Copyright Marcus Britanicus <marcusbritanicus@gmail.com>
 *
 * This file is a part of LibDesQ (https://gitlab.com/DesQ/libdesq)
 * This library contains various  Core and Gui classes which are used
 * through out the DesQ Project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * at your option, any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 **/

#include <stdio.h>
#include <stdlib.h>

#include "DesQUtils.hpp"
#include "desq-config.h"

static QMimeDatabase mimeDb;
FILE                 *DesQ::log = nullptr;
bool                 DesQ::SHOW_INFO_ON_CONSOLE     = true;
bool                 DesQ::SHOW_DEBUG_ON_CONSOLE    = true;
bool                 DesQ::SHOW_WARNING_ON_CONSOLE  = true;
bool                 DesQ::SHOW_CRITICAL_ON_CONSOLE = true;
bool                 DesQ::SHOW_FATAL_ON_CONSOLE    = true;

static const char *COLOR_INFO     = "\033[01;32m";
static const char *COLOR_DEBUG    = "\033[01;30m";
static const char *COLOR_WARN     = "\033[01;33m";
static const char *COLOR_CRITICAL = "\033[01;31m";
static const char *COLOR_FATAL    = "\033[01;41m";
static const char *COLOR_RESET    = "\033[00;00m";

void Logger( QtMsgType type, const QMessageLogContext& context, const QString& msg ) {
    QByteArray localMsg  = msg.toLocal8Bit();
    const char *file     = context.file ? context.file : "";
    const char *function = context.function ? context.function : "";

    switch ( type ) {
        case QtInfoMsg: {
            if ( DesQ::SHOW_INFO_ON_CONSOLE ) {
                fprintf( stderr, "%s[I]: (%s:%u, %s) %s%s\n",
                         COLOR_INFO,
                         file,
                         context.line,
                         function,
                         localMsg.constData(),
                         COLOR_RESET
                );
                fflush( stderr );
            }

            if ( DesQ::log ) {
                fprintf( DesQ::log, "[I]: (%s:%u, %s) %s\n", file, context.line, function, localMsg.constData() );
                fflush( DesQ::log );
            }

            break;
        }

        case QtDebugMsg: {
            if ( DesQ::SHOW_DEBUG_ON_CONSOLE ) {
                fprintf( stderr, "%s[D]: (%s:%u, %s) %s%s\n",
                         COLOR_DEBUG,
                         file,
                         context.line,
                         function,
                         localMsg.constData(),
                         COLOR_RESET
                );
                fflush( stderr );
            }

            if ( DesQ::log ) {
                fprintf( DesQ::log, "[D]: (%s:%u, %s) %s\n", file, context.line, function, localMsg.constData() );
                fflush( DesQ::log );
            }

            break;
        }

        case QtWarningMsg: {
            if ( DesQ::SHOW_WARNING_ON_CONSOLE ) {
                fprintf( stderr, "%s[W]: (%s:%u, %s) %s%s\n",
                         COLOR_WARN,
                         file,
                         context.line,
                         function,
                         localMsg.constData(),
                         COLOR_RESET
                );
                fflush( stderr );
            }

            if ( DesQ::log ) {
                fprintf( DesQ::log, "[W]: (%s:%u, %s) %s\n", file, context.line, function, localMsg.constData() );
                fflush( DesQ::log );
            }

            break;
        }

        case QtCriticalMsg: {
            if ( DesQ::SHOW_CRITICAL_ON_CONSOLE ) {
                fprintf( stderr, "%s[E]: (%s:%u, %s) %s%s\n",
                         COLOR_CRITICAL,
                         file,
                         context.line,
                         function,
                         localMsg.constData(),
                         COLOR_RESET
                );
                fflush( stderr );
            }

            if ( DesQ::log ) {
                fprintf( DesQ::log, "[E]: (%s:%u, %s) %s\n", file, context.line, function, localMsg.constData() );
                fflush( DesQ::log );
            }

            break;
        }

        case QtFatalMsg: {
            if ( DesQ::SHOW_FATAL_ON_CONSOLE ) {
                fprintf( stderr, "%s[#]: (%s:%u, %s) %s%s\n",
                         COLOR_FATAL,
                         file,
                         context.line,
                         function,
                         localMsg.constData(),
                         COLOR_RESET
                );
                fflush( stderr );
            }

            if ( DesQ::log ) {
                fprintf( DesQ::log, "[#]: (%s:%u, %s) %s\n", file, context.line, function, localMsg.constData() );
                fflush( DesQ::log );
            }

            break;
        }
    }
}


QString DesQ::Utils::dirName( QString path ) {
    if ( path == "/" or path == "//" ) {
        return "/";
    }

    /* Simple clean path: remove '//' and './' */
    path = path.replace( "//", "/" ).replace( "/./", "/" );

    char    *dupPath = strdup( path.toLocal8Bit().constData() );
    QString dirPth   = QString::fromLocal8Bit( dirname( dupPath ) );

    dirPth += (dirPth.endsWith( "/" ) ? "" : "/");
    free( dupPath );

    return dirPth;
}


QString DesQ::Utils::baseName( QString path ) {
    if ( path == "/" or path == "//" ) {
        return "/";
    }

    /* Simple clean path" remove '//' and './' */
    path = path.replace( "//", "/" ).replace( "/./", "/" );

    char    *dupPath = strdup( path.toLocal8Bit().constData() );
    QString basePth  = QString::fromLocal8Bit( basename( dupPath ) );

    free( dupPath );

    return basePth;
}


QString DesQ::Utils::getMimeType( QString path ) {
    return mimeDb.mimeTypeForFile( path ).name();
}


QString DesQ::Utils::getMimeIcon( QString path ) {
    return mimeDb.mimeTypeForFile( path ).iconName();
}


bool DesQ::Utils::isFile( QString path ) {
    struct stat statbuf;

    if ( stat( path.toLocal8Bit().constData(), &statbuf ) == 0 ) {
        if ( S_ISREG( statbuf.st_mode ) or S_ISLNK( statbuf.st_mode ) ) {
            return true;
        }

        else {
            return false;
        }
    }

    else {
        return false;
    }
}


bool DesQ::Utils::isDir( QString path ) {
    struct stat statbuf;

    if ( stat( path.toLocal8Bit().constData(), &statbuf ) == 0 ) {
        if ( S_ISDIR( statbuf.st_mode ) ) {
            return true;
        }

        else {
            return false;
        }
    }

    else {
        return false;
    }
}


bool DesQ::Utils::isLink( QString path ) {
    struct stat statbuf;

    if ( lstat( path.toLocal8Bit().constData(), &statbuf ) == 0 ) {
        if ( S_ISLNK( statbuf.st_mode ) ) {
            return true;
        }

        else {
            return false;
        }
    }

    else {
        return false;
    }
}


bool DesQ::Utils::exists( QString path ) {
    return not access( path.toLocal8Bit().constData(), F_OK );
}


QString DesQ::Utils::readLink( QString path ) {
    char linkTarget[ PATH_MAX ] = { 0 };

    realpath( path.toLocal8Bit().constData(), linkTarget );

    return QString( linkTarget );
}


int DesQ::Utils::mkpath( QString path, mode_t mode ) {
    /* Root always exists */
    if ( path == "/" ) {
        return 0;
    }

    /* If the directory exists, thats okay for us */
    if ( DesQ::Utils::exists( path ) ) {
        return 0;
    }

    mkpath( DesQ::Utils::dirName( path ), mode );

    return mkdir( path.toLocal8Bit().constData(), mode );
}


bool DesQ::Utils::removeDir( QString dirName ) {
    bool result = true;
    QDir dir( dirName );

    QDirIterator it( dirName, QDir::NoDotAndDotDot | QDir::System | QDir::Hidden | QDir::AllDirs | QDir::Files );

    while ( it.hasNext() ) {
        it.next();

        if ( it.fileInfo().isDir() ) {
            result &= DesQ::Utils::removeDir( it.filePath() );
        }

        else {
            result &= QFile::remove( it.filePath() );
        }
    }

    rmdir( dirName.toUtf8().constData() );

    return result;
}


bool DesQ::Utils::isReadable( QString path ) {
    if ( isDir( path ) ) {
        return not access( path.toLocal8Bit().constData(), R_OK | X_OK );
    }

    else {
        return not access( path.toLocal8Bit().constData(), R_OK );
    }
}


bool DesQ::Utils::isWritable( QString path ) {
    return not access( path.toLocal8Bit().constData(), W_OK );
}


bool DesQ::Utils::isExecutable( QString path ) {
    struct stat statbuf;

    if ( stat( path.toLocal8Bit().data(), &statbuf ) != 0 ) {
        return false;
    }

    if ( (statbuf.st_mode & S_IXUSR) ) {
        QMimeType m = mimeDb.mimeTypeForFile( path );

        if ( m.name() == "application/x-executable" ) {
            return true;
        }

        else if ( m.name() == "application/x-sharedlib" ) {
            return true;
        }

        else if ( m.allAncestors().contains( "application/x-executable" ) ) {
            return true;
        }

        /* Default is false */
        return false;
    }

    return false;
}


qint64 DesQ::Utils::getSize( QString path ) {
    struct stat statbuf;

    if ( stat( path.toLocal8Bit().constData(), &statbuf ) != 0 ) {
        return 0;
    }

    switch ( statbuf.st_mode & S_IFMT ) {
        case S_IFREG: {
            return statbuf.st_size;
        }

        case S_IFDIR: {
            DIR           *d_fh;
            struct dirent *entry;
            QString       longest_name;

            while ( (d_fh = opendir( path.toLocal8Bit().constData() ) ) == NULL ) {
                qWarning() << "Couldn't open directory:" << path;
                return statbuf.st_size;
            }

            qint64 size = statbuf.st_size;

            longest_name = QString( path );

            if ( not longest_name.endsWith( "/" ) ) {
                longest_name += "/";
            }

            while ( (entry = readdir( d_fh ) ) != NULL ) {
                /* Don't ascend up the tree or include the current directory */
                if ( (strcmp( entry->d_name, ".." ) != 0) && (strcmp( entry->d_name, "." ) != 0) ) {
                    if ( entry->d_type == DT_DIR ) {
                        /* Recurse into that folder */
                        size += getSize( longest_name + entry->d_name );
                    }

                    else {
                        /* Get the size of the current file */
                        size += getSize( longest_name + entry->d_name );
                    }
                }
            }

            closedir( d_fh );
            return size;
        }

        default: {
            /* Return 0 for all other nodes: chr, blk, lnk, symlink etc */
            return 0;
        }
    }

    /* Should never come till here */
    return 0;
}


mode_t DesQ::Utils::getMode( QString path ) {
    struct stat fileAtts;

    if ( stat( path.toLocal8Bit().constData(), &fileAtts ) != 0 ) {
        return -1;
    }

    return fileAtts.st_mode;
}


QStringList DesQ::Utils::recDirWalk( QString path ) {
    QStringList fileList;

    if ( not QFileInfo( path ).exists() ) {
        return QStringList();
    }

    QDirIterator it( path, QDir::AllEntries | QDir::System | QDir::NoDotAndDotDot | QDir::Hidden, QDirIterator::Subdirectories );

    while ( it.hasNext() ) {
        it.next();

        if ( it.fileInfo().isDir() ) {
            fileList.append( it.fileInfo().filePath() );
        }

        else {
            fileList.append( it.fileInfo().filePath() );
        }
    }

    return fileList;
}


QString DesQ::Utils::formatSize( qint64 size, qint64 hint ) {
    if ( hint == SizeHint::Auto ) {
        if ( size >= SizeHint::TiB ) {
            return QString( "%1 TiB" ).arg( QString::number( qreal( size ) / SizeHint::TiB, 'f', 3 ) );
        }
        else if ( size >= SizeHint::GiB ) {
            return QString( "%1 GiB" ).arg( QString::number( qreal( size ) / SizeHint::GiB, 'f', 2 ) );
        }
        else if ( size >= SizeHint::MiB ) {
            return QString( "%1 MiB" ).arg( QString::number( qreal( size ) / SizeHint::MiB, 'f', 1 ) );
        }
        else if ( size >= SizeHint::KiB ) {
            return QString( "%1 KiB" ).arg( QString::number( qreal( size ) / SizeHint::KiB, 'f', 1 ) );
        }
        else {
            return QString( "%1 B" ).arg( size );
        }
    }

    else {
        if ( hint == SizeHint::TiB ) {
            return QString( "%1 TiB" ).arg( QString::number( qreal( size ) / SizeHint::TiB, 'f', 3 ) );
        }
        else if ( hint == SizeHint::GiB ) {
            return QString( "%1 GiB" ).arg( QString::number( qreal( size ) / SizeHint::GiB, 'f', 2 ) );
        }
        else if ( hint == SizeHint::MiB ) {
            return QString( "%1 MiB" ).arg( QString::number( qreal( size ) / SizeHint::MiB, 'f', 1 ) );
        }
        else if ( hint == SizeHint::KiB ) {
            return QString( "%1 KiB" ).arg( QString::number( qreal( size ) / SizeHint::KiB, 'f', 1 ) );
        }
        else {
            return QString( "%1 B" ).arg( size );
        }
    }

    return QString();
}


double DesQ::Utils::formatSizeRaw( qint64 size, qint64 hint ) {
    if ( hint == SizeHint::Auto ) {
        if ( size >= SizeHint::TiB ) {
            return qreal( size ) / SizeHint::TiB;
        }
        else if ( size >= SizeHint::GiB ) {
            return qreal( size ) / SizeHint::GiB;
        }
        else if ( size >= SizeHint::MiB ) {
            return qreal( size ) / SizeHint::MiB;
        }
        else if ( size >= SizeHint::KiB ) {
            return qreal( size ) / SizeHint::KiB;
        }
        else {
            return size;
        }
    }

    else {
        if ( hint == SizeHint::TiB ) {
            return qreal( size ) / SizeHint::TiB;
        }
        else if ( hint == SizeHint::GiB ) {
            return qreal( size ) / SizeHint::GiB;
        }
        else if ( hint == SizeHint::MiB ) {
            return qreal( size ) / SizeHint::MiB;
        }
        else if ( hint == SizeHint::KiB ) {
            return qreal( size ) / SizeHint::KiB;
        }
        else {
            return size;
        }
    }

    return size;
}


QString DesQ::Utils::formatSizeStr( qint64 size, qint64 hint ) {
    if ( hint == SizeHint::Auto ) {
        if ( size >= SizeHint::TiB ) {
            return QString( "TiB" );
        }
        else if ( size >= SizeHint::GiB ) {
            return QString( "GiB" );
        }
        else if ( size >= SizeHint::MiB ) {
            return QString( "MiB" );
        }
        else if ( size >= SizeHint::KiB ) {
            return QString( "KiB" );
        }
        else {
            return QString( "B" );
        }
    }

    else {
        if ( hint == SizeHint::TiB ) {
            return QString( "TiB" );
        }
        else if ( hint == SizeHint::GiB ) {
            return QString( "GiB" );
        }
        else if ( hint == SizeHint::MiB ) {
            return QString( "MiB" );
        }
        else if ( hint == SizeHint::KiB ) {
            return QString( "KiB" );
        }
        else {
            return QString( "B" );
        }
    }

    return "B";
}


QString DesQ::Utils::getUtilityPath( QString utility ) {
    return UtilsPath "desq-" + utility;
}
